import base64
from . import core
from typing import Iterable


class Post(core.Post):
    def __init__(self):
        self._index = 0
        self._source = ""
        self._timestamp = 0.0
        self._replying_to = []
        self._tags = []
        self._m_retrieved = False
        self._metadata = b""
        self._d_retrieved = False
        self._data = b""
        self._encryption = "aes-gcm"
        self.bb = core.Blackboard()

    # TODO: This may need to do error handling
    # also, how to return a concrete Post with type hints?
    @classmethod
    def from_dict(cls, d: dict):
        p = Post()
        p._index = d["index"]
        p._source = d["source"]
        p._timestamp = d["timestamp"]
        p._replying_to = d["replying_to"]
        p._tags = d["tags"]
        if d["metadata"] is not None:
            p._metadata = base64.b64decode(d["metadata"])
        if d["data"] is not None:
            p._data = base64.b64decode(d["data"])
        p._encryption = d["encryption"]
        return p

    # Access only (set by the server)
    def index(self) -> int:
        return self._index

    def set_index(self, idx: int):
        self._index = idx

    def timestamp(self) -> float:
        return self._timestamp

    # Read/Write
    def source(self) -> str:
        return self._source

    def set_source(self, src: str):
        self._source = src

    def tags(self) -> Iterable[str]:
        return self._tags

    def set_tags(self, tags: Iterable[str]):
        self._tags = tags

    def replying_to(self) -> Iterable[int]:
        return self._replying_to

    def set_replying_to(self, replying_to: Iterable[int]):
        self._replying_to = replying_to

    async def metadata(self) -> bytes | Exception:
        if len(self._metadata) == 0 or self._m_retrieved:
            return self._metadata

        # Retrieve the underlying data using the embedded transit
        obj_id = self._metadata.decode()
        metadata_or_err = await self.bb.get_obj(obj_id)
        if isinstance(metadata_or_err, Exception):
            return metadata_or_err

        self._metadata = metadata_or_err
        self._m_retrieved = True

        return self._metadata

    def set_metadata(self, meta: bytes):
        self._metadata = meta
        self._m_retrieved = True

    async def data(self) -> bytes | Exception:
        if len(self._data) == 0 or self._d_retrieved:
            return self._data
        #
        # Retrieve the underlying data using the embedded transit
        obj_id = self._data.decode()
        data_or_err = await self.bb.get_obj(obj_id)
        if isinstance(data_or_err, Exception):
            return data_or_err

        self._data = data_or_err
        self._d_retrieved = True

        return self._data

    def set_data(self, data: bytes):
        self._data = data
        self._d_retrieved = True

    def encryption(self) -> str:
        return self._encryption

    def set_encryption(self, enc: str):
        self._encryption = enc



import asyncio
from typing import Iterable
from . import core
from .tag_set import TagSet


class Filter:
    async def filter(self, post: core.Post) -> None | Exception:
        raise NotImplementedError

    def outgoing(self) -> asyncio.Queue:
        raise NotImplementedError


class TagFilter(Filter):
    def __init__(self, tags: Iterable[str]):
        self.tags = tags
        self.tag_set = TagSet(tags)
        self.outgoing_ch = asyncio.Queue()

    # In the go implementation we return an error if the filter
    # does NOT match the incoming post.  There's no similar
    # 1 to 1 code replacement here so we use the union trick that
    # is meh
    async def filter(self, post: core.Post) -> None | Exception:
        incoming_tags = TagSet(post.tags())

        if not self.tag_set.matches(incoming_tags):
            return Exception("message was filtered")

        await self.outgoing_ch.put(post)

    def outgoing(self) -> asyncio.Queue:
        return self.outgoing_ch


class FilterBank:
    def __init__(self, filters=None):
        self.lock = asyncio.Lock()
        self.filters = []
        if filters is not None:
            self.filters = filters

    async def len(self) -> int:
        async with self.lock:
            return len(self.filters)

    async def insert(self, filt: Filter):
        async with self.lock:
            self.filters.append(filt)

    # Run a post through the filter bank, populating
    # the outgoing channels that match
    async def filter(self, post: core.Post):
        async with self.lock:
            for filt in self.filters:
                await filt.filter(post)



from typing import Iterable
from . import core
from . import filter

import asyncio
# Our default core agent


# Agents assume zero indexing.  It is the responsiblity
# of the specific *blackboard* implementation to handle any translation.
class Agent:
    def __init__(self, bb: core.Blackboard):
        self.last_read = -1
        self.bb = bb  # blackboard for the agent to talk to
        self.filters = filter.FilterBank()

    async def fetch_loop(self):
        while True:
            await asyncio.sleep(0.25)

            err = await self.fetch()
            if isinstance(err, Exception):
                return err

    def start(self) -> asyncio.Task:
        return asyncio.create_task(self.fetch_loop())

    # Fetch primes the filterbank with any new posts
    async def fetch(self) -> None | Exception:
        bb_len = await self.bb.len()
        if isinstance(bb_len, Exception):
            return bb_len

        if self.last_read + 1 == bb_len or bb_len == 0:
            return

        for i in range(self.last_read + 1, bb_len):
            post = await self.bb.read(i)

            if isinstance(post, Exception):
                return post

            await self.filters.filter(post)
            self.last_read = i

    async def listen_for(self, *tags) -> asyncio.Queue:
        tf = filter.TagFilter(*tags)
        await self.filters.insert(tf)
        return tf.outgoing()

    async def hello(self) -> None | Exception:
        post = self.bb.message_hello()
        return await self.bb.write(post)

    async def goodbye(self) -> None | Exception:
        post = self.bb.message_goodbye()
        return await self.bb.write(post)

    async def post(
        self, metadata: bytes | None, data: bytes | None, tags: Iterable[str]
    ) -> None | Exception:
        # Get a post structure from the underlying transit
        post = self.bb.new_post()

        if metadata is not None:
            post.set_metadata(metadata)
        if data is not None:
            post.set_data(data)

        post.set_tags(tags)
        return await self.bb.write(post)

    async def reply(
        self,
        posts: Iterable[core.Post],
        metadata: bytes | None,
        data: bytes | None,
        tags: Iterable[str],
    ) -> None | Exception:
        # Get a post structure from the underlying transit
        new_post = self.bb.new_post()

        reply_to = []
        for post in posts:
            reply_to.append(post.index())

        new_post.set_replying_to(reply_to)

        if metadata is not None:
            new_post.set_metadata(metadata)
        if data is not None:
            new_post.set_data(data)
        new_post.set_tags(tags)

        return await self.bb.write(new_post)



from . import core
from .hardcore_post import Post
import json
import asyncio
import aiohttp
import base64

# John: you have some clarifications to make around what the input to this
# layer is and whether or not we case use knowledge of the underlying
# post type here.  I think if we do, it kind of undermines the entire
# premise of breaking this into a separate layer, but there's also
# no way around it in something like the new_post method: the bb has
# to have some awareness of what it's talking to, but only minimally (I think
# why I'm having trouble answering this question in the first place).


class HTTPTransit(core.Blackboard):
    read_path = "/read/{0}"
    write_path = "/write"
    len_path = "/len"
    put_obj_path = "/obj"
    get_obj_path = "/obj/{0}"

    def __init__(self, bb_addr: str):
        self._name = "http-transit"
        self.addr = bb_addr
        self.session = aiohttp.ClientSession()

    def __del__(self):
        """
        Destructor for BlackboardHTTPClient class.
        """
        asyncio.create_task(self.session.close())

    # A little bit of future-proofing to allow for easily changed common url
    # components
    def _get_url(self, path: str):
        return "http://" + self.addr + path

    # Our transit layers need to be identifiable
    def name(self) -> str:
        return self._name

    def set_name(self, name: str):
        self._name = name

    # This is core-specific means of handling data and metadata.
    # The client/agent pushes data into the object store first
    # and then passes on the link in the final message to the BB.
    async def put_obj(self, data: bytes) -> str | Exception:
        url = self._get_url(self.put_obj_path)

        async with self.session.post(url, data=data) as response:
            if response.status != 200:
                return RuntimeError("object upload failed")

            # form the url other agents can use to retrieve the data
            oid = await response.text()
            get_url = self._get_url(self.get_obj_path.format(oid))
            return get_url

    # For an http blackboard, the id is a url on which we can call "get"
    async def get_obj(self, id: str) -> bytes | Exception:
        async with self.session.get(id) as response:
            if response.status != 200:
                return RuntimeError("object upload failed")

            return await response.content.read()

    # Fulfill the required NotImplementedError class methods
    async def read(self, idx: int) -> core.Post | Exception:
        url = self._get_url(self.read_path.format(idx))

        async with self.session.get(url) as response:
            if response.status != 200:
                # TODO: This exc probably not complete enough to be very useful
                msg = await response.text()
                status_message = f"{response.status} {msg}"
                return RuntimeError(f"read failed; status: {status_message}")

            body = await response.text()
            msg = json.loads(body)

            # Unwrap the json dictionary into a post instance
            # Data and metadata come to us as base64 encoded strings
            # p = self.new_post()

            p = Post.from_dict(msg)
            p.bb = self

            return p

    async def write(self, post: core.Post) -> None | Exception:
        # TODO: encryption should be enforced here
        url = self._get_url(self.write_path)

        post_dict = {}
        post_dict["index"] = post.index()
        post_dict["source"] = post.source()
        post_dict["timestamp"] = post.timestamp()
        post_dict["tags"] = post.tags()
        post_dict["replying_to"] = post.replying_to()
        # post_dict["metadata"] = base64.b64encode(metadata_url).decode()
        # post_dict["data"] = base64.b64encode(data).decode()
        post_dict["encryption"] = post.encryption()

        # Perform our data stripping on the client side
        data = await post.data()
        if isinstance(data, Exception):
            err = data
            return err

        if len(data) != 0:
            data_url = await self.put_obj(data)
            if isinstance(data_url, Exception):
                return data_url
            post_dict["data"] = base64.b64encode(data_url.encode()).decode()

        # Perform our data stripping on the client side
        m_data = await post.metadata()
        if isinstance(m_data, Exception):
            err = m_data
            return err

        if len(m_data) != 0:
            metadata_url = await self.put_obj(m_data)
            if isinstance(metadata_url, Exception):
                return metadata_url
            post_dict["metadata"] = base64.b64encode(
                metadata_url.encode()).decode()

        # sigh
        msg_json = json.dumps(post_dict)
        payload = msg_json.encode("utf-8")

        async with self.session.post(
            url,
            data=payload,
            headers={
                "content-type": "application/json",
                "content-length": str(len(payload)),
            },
        ) as response:
            if response.status != 201:
                msg = await response.text()
                status_message = f"{response.status} {msg}"
                return RuntimeError(f"write failed; status: {status_message}")

    async def len(self) -> int | Exception:
        url = self._get_url(self.len_path)

        async with self.session.get(url) as response:
            if response.status != 200:
                msg = await response.text()
                status_message = f"{response.status} {msg}"
                return RuntimeError(f"len failed; status: {status_message}")

            body = await response.text()
            len_str = body.strip()
            if not len_str.isdigit():
                return RuntimeError(f"non-numeric length response: {len_str}")

            return int(len_str)

    def new_post(self) -> core.Post:
        p = Post()
        p.bb = self
        p.set_source(self._name)
        return p



import io
import numpy as np
import time
import zlib
import tarfile
import glob
import os
import json
import base64


def attributes(attrs: dict) -> bytes:
    return base64.b64encode(json.dumps(attrs).encode("utf-8"))


def dictionary(input: dict) -> bytes:
    return json.dumps(input).encode("utf-8")


def file(path) -> bytes:
    """
    file serializes a file on disk to binary.

    Parameters
    ----------
    path (str): path to the file.

    Returns
    -------
    bytes object containing serialized data.
    """
    with open(path, "rb") as f:
        return f.read()


def compressed_file(path: str) -> bytes:
    """
    compressed_file compresses and serializes a file on disk to binary.

    Parameters
    ----------
    path (str): path to the file.

    Returns
    -------
    bytes object containing serialized data.
    """
    with open(path, "rb") as f:
        tic = time.perf_counter()
        data = f.read()
        compressed_data = zlib.compress(data)
        compression_factor = float(len(compressed_data)) / float(len(data))
        print(
            f"compressed size: {100 * compression_factor:2.0f}% of original in {1000 * (time.perf_counter() - tic):2.0f} ms"
        )
        return compressed_data


# Numpy array, no compression
def numpy(array):
    with io.BytesIO() as bytes_buffer:
        np.save(bytes_buffer, array, allow_pickle=True)
        bytes_buffer.seek(0)
        return bytes_buffer.read()


# Numpy array, with compression (use for segmentations/masks)
def compressed_numpy(array):
    with io.BytesIO() as bytes_buffer:
        np.savez_compressed(bytes_buffer, array=array)
        bytes_buffer.seek(0)
        return bytes_buffer.read()


# We have found the concept of a dicom dir or a dicom dicom_series
# to be a generally useful unit of operation.LL
def directory(dir_path):
    with io.BytesIO() as bytes_buffer:
        with tarfile.open(None, "w:gz", fileobj=bytes_buffer) as tar:
            tar.add(dir_path, arcname=os.path.basename(dir_path))

        bytes_buffer.seek(0)
        return bytes_buffer.read()


# NOTE: the agent context serialization deserialization may
# ultimately belong elsewhere.
def agent_context(dir_path: str) -> bytes:
    """
    Pack context contained in dir_path. Return
    as bytes.
    """
    packed_data = io.BytesIO()
    with tarfile.open(mode="w:gz", fileobj=packed_data) as f:
        for fp in glob.glob(dir_path + "/*"):
            arcname = os.path.relpath(fp, dir_path)
            f.add(fp, arcname=arcname)

    packed_data.seek(0, os.SEEK_SET)
    return packed_data.read()


# if __name__ == "__main__":
#     test_data = pack_agent_context("test-dir")
#     unpack_agent_context(test_data, "./output-unpack")



# Module providing the post and transit layers for an agent based on the
# *hardcore protocol*.  The goal is a simpler agent implementation that still
# preserves all current functionality on previous version of core.

from .hardcore_post import Post
from .http_transit import HTTPTransit
from .sqlite_transit import SQLiteTransit
from .postgres_transit import PostgreSQLTransit

# Provide an implementation of Post that is compatible with Core



import os
import argparse
import sys
import asyncio
import random
from . import hardcore
from . import agent

global bb_addr
bb_addr = "localhost:8080"


class bcolors:
    OKRED = "\033[31m"
    OKGREEN = "\033[32m"
    OKYELLOW = "\033[33m"
    OKBLUE = "\033[34m"
    OKMAGENTA = "\033[35m"
    OKCYAN = "\033[36m"
    OKBBLUE = "\033[94m"
    OKBCYAN = "\033[96m"
    OKBGREEN = "\033[92m"
    ENDC = "\033[0m"
    UNDERLINE = "\033[4m"


source_colors = {
    bcolors.OKRED,
    bcolors.OKGREEN,
    bcolors.OKYELLOW,
    bcolors.OKBLUE,
    bcolors.OKMAGENTA,
    bcolors.OKCYAN,
    bcolors.OKBBLUE,
    bcolors.OKBCYAN,
    bcolors.OKBGREEN,
}


# class bcolors:
#     HEADER = '\033[95m'
#     OKBLUE = '\033[94m'
#     OKCYAN = '\033[96m'
#     OKGREEN = '\033[92m'
#     WARNING = '\033[93m'
#     FAIL = '\033[91m'
#     ENDC = '\033[0m'
#     BOLD = '\033[1m'
#     UNDERLINE = '\033[4m'


async def main():
    if bb_addr is None or bb_addr == "":
        print("Error: invalid BB address")
        sys.exit(1)
    else:
        print(f"blackboard address is: {bb_addr}")

    if bb_addr.endswith(".db") or bb_addr.endswith(".sqlite"):
        transit = hardcore.SQLiteTransit(bb_addr)
    else:
        transit = hardcore.HTTPTransit(bb_addr)

    bb = transit
    a = agent.Agent(bb)
    a.start()

    ch = await a.listen_for([])
    sources = {}
    while True:
        # await asyncio.sleep(1.0)
        # print(await a.bb.len())

        # TODO: some light formatting of the post for nicer display
        p = await ch.get()

        if p.source() not in sources:
            sources[p.source()] = random.choice(list(source_colors))

        print(
            f"{p.index():08d} {sources[p.source()]}{p.source():24s}{bcolors.ENDC} {str(p.replying_to()):8s} {[tag[0:24] for tag in p.tags()]} "
        )


if __name__ == "__main__":
    ap = argparse.ArgumentParser(prog="smcore.listen")
    ap.add_argument(
        "--addr",
        default="localhost:8080",
        help="set the blackboard address",
    )
    args = ap.parse_args()
    bb_addr = args.addr

    asyncio.run(main())



import io

# from tempfile import TemporaryDirectory
import tempfile
import numpy as np
import zlib
import tarfile
import os
import json
import base64


def attributes(data: bytes) -> dict:
    return json.loads(base64.b64decode(data).decode("utf-8"))


def dictionary(data: bytes) -> dict:
    return json.loads(data.decode("utf-8"))


def file(data: bytes, save_path: str):
    """
    file writes binary data to disk. File-based deserialization
    can be useful if a library doesn't present memory-level interface.

    Parameters
    ----------
    data (bytes): serialized data to write to a file
    save_path (str): path to write the file to.
    """
    with open(save_path, "wb") as f:
        f.write(data)


def compressed_file(data: bytes, save_path: str):
    """
    compressed_file decompresses binary data using zlib and saves
    it to disk.

    Parameters
    ----------
    data (bytes): serialized data to write to a file.
    save_path (str): path to write the file to.
    """
    data = zlib.decompress(data)
    with open(save_path, "wb") as f:
        f.write(data)


def compressed_memory_file(data: bytes):
    """
    compressed_memory_file decompresses binary data using zlib and
    saves it to memory as an io.BytesIO object.

    Parameters
    ----------
    data (bytes): serialized data to decompress and write to memory.

    Returns
    -------
    Decompressed,
    """
    return io.BytesIO(zlib.decompress(data))


def memory_file(data: bytes) -> io.BytesIO:
    """
    memory_file saves binary data to memory. BytesIO behaves like
    file, but exists in memory. This is more efficient and recommended
    if it works in your use case.

    Parameters
    ----------
    data (bytes): serialized data to write to memory.

    Returns
    -------
    BytesIO-encoded data.
    """
    return io.BytesIO(data)


def numpy(data: bytes) -> np.ndarray[float]:
    """
    numpy deserializes binary data to a numpy array.

    Parameters
    ----------
    data (bytes): serialized data to write to an array.

    Returns
    -------
    np.ndarray[float] object with deserialized data.
    """
    buff = io.BytesIO(data)
    return np.load(buff, allow_pickle=True)


def compressed_numpy(data: bytes):
    """
    compressed_numpy deserializes binary data to a numpy array.
    It is the same function as numpy, but we include it to maintain
    a consistent interface.

    Parameters
    ----------
    data (bytes): serialized data to write to an array.

    Returns
    -------
    np.ndarray[float] object with deserialized data.
    """
    buff = io.BytesIO(data)
    loaded = np.load(buff, allow_pickle=True)
    return loaded["array"]


# WARN: responsibility for cleanup falls to whoever uses 'str'; could fill up some disks
def directory(data: bytes) -> str:
    tmp = tempfile.mkdtemp()

    buff = io.BytesIO(data)
    with tarfile.open(None, "r:gz", fileobj=buff) as tar:
        tar.extractall(path=tmp, filter="data")

    return tmp


# NOTE: the agent context serialization deserialization may
# ultimately belong elsewhere.
def agent_context(data: bytes, dir_path: str):
    """
    Unpack context contained in 'data' into the specified directory
    dir_path. dir_path will be created if it does not exist.
    """
    packed_data = io.BytesIO(data)
    packed_data.seek(0, os.SEEK_SET)

    with tarfile.open(mode="r:gz", fileobj=packed_data) as f:
        if not os.path.exists(dir_path):
            os.makedirs(dir_path, mode=0o775, exist_ok=True)
        f.extractall(path=dir_path)



import os
import sqlite3
from . import core
from .hardcore_post import Post

# SQLite transit is built on top of hardcore posts
#
# I haven't vetted this for security, in particular SQL injections.  I've tried
# my best to follow the advice given in the docs for sqlite3.


class SQLiteTransit(core.Blackboard):
    def __init__(self, path: str):
        self._name = "sqlite-transit"
        self.con = sqlite3.connect(path)
        self.cur = self.con.cursor()

        # provision the db if needed
        # "index" is "id" since index is a keyword
        self.cur.execute("""
            CREATE TABLE IF NOT EXISTS posts(
                idx INTEGER PRIMARY KEY,
                source TEXT,
                timestamp DATETIME DEFAULT CURRENT_TIMESTAMP,
                metadata TEXT,
                data TEXT,
                encryption TEXT DEFAULT 'aes-gcm'
            );
            """)
        self.con.commit()

        self.cur.execute("CREATE TABLE IF NOT EXISTS tags(tag TEXT UNIQUE);")
        self.con.commit()

        self.cur.execute(
            "CREATE TABLE IF NOT EXISTS post_tags(post INTEGER, tag TEXT, FOREIGN KEY(post) REFERENCES posts(idx),FOREIGN KEY(tag) REFERENCES tags(tag))"
        )
        self.con.commit()

        self.cur.execute("""
            CREATE TABLE IF NOT EXISTS replies(
                post INTEGER,
                replying_to INTEGER,
                FOREIGN KEY(post) REFERENCES posts(idx),
                FOREIGN KEY(replying_to) REFERENCES posts(idx)
            );
            """)
        self.con.commit()

        # Ensure we can use a directory next to the DB for an object store
        db_dir = os.path.dirname(path)
        self.obj_dir = os.path.join(db_dir, "obj")
        try:
            os.makedirs(self.obj_dir)
        except:
            pass

    def name(self) -> str:
        return self._name

    def set_name(self, name: str):
        self._name = name

    async def put_obj(self, data: bytes) -> str | Exception:
        obj_id = core.random_id()
        obj_path = os.path.join(self.obj_dir, obj_id)
        with open(obj_path, "wb") as f:
            f.write(data)
        return obj_id

    async def get_obj(self, id: str) -> bytes | Exception:
        obj_path = os.path.join(self.obj_dir, id)
        with open(obj_path, "rb") as f:
            return f.read()

    # NOTE: Agents are expecting to index from zero, but SQLite starts with row idx 1
    async def read(self, idx: int) -> core.Post | Exception:
        adjusted_idx = idx + 1
        result = self.cur.execute(
            "SELECT idx,source,metadata,data FROM posts WHERE idx IS ?", (
                adjusted_idx,)
        )
        self.con.commit()

        post_data = result.fetchone()  # This should be unique
        if post_data is None:
            return RuntimeError(f"idx {adjusted_idx} not found")

        # Query the tags associated with the post
        result = self.cur.execute(
            "SELECT tag FROM post_tags WHERE post IS ?", (adjusted_idx,)
        )
        self.con.commit()

        tags = result.fetchall()  # This should be unique
        tag_list = [x[0] for x in tags]

        # Query replies
        result = self.cur.execute(
            "SELECT replying_to FROM replies WHERE post IS ?", (adjusted_idx,)
        )
        self.con.commit()

        replies = result.fetchall()  # This should be unique
        reply_list = [x[0] for x in replies]

        # argument for: agents should never really be dealing with those
        # indices directly. against: we expose the index to the user via the
        # post.index() method, if this
        # is not internally consistent, it will make for bad behavior.
        post = self.new_post()
        # this will return an index to the agent that they could use to 'read(idx)' to get the message back.
        post.set_index(post_data[0] - 1)
        post.set_source(post_data[1])
        post.set_tags(tag_list)
        post.set_metadata(post_data[2].encode())
        post.set_data(post_data[3].encode())
        post.set_replying_to(reply_list)

        # temporary kluge
        post._m_retrieved = False
        post._d_retrieved = False

        return post

    async def write(self, post: core.Post) -> None | Exception:
        # Indirect our data onto disk

        m_data = await post.metadata()
        if isinstance(m_data, Exception):
            return m_data

        metadata_obj_id = ""
        if len(m_data) > 0:
            metadata_obj_id = await self.put_obj(m_data)
            if isinstance(metadata_obj_id, Exception):
                return metadata_obj_id

        data = await post.data()
        if isinstance(data, Exception):
            return data

        data_obj_id = ""
        if len(data) > 0:
            data_obj_id = await self.put_obj(data)
            if isinstance(data_obj_id, Exception):
                return data_obj_id
            post.set_data(data_obj_id.encode())

        ins_post = (
            post.source(),
            metadata_obj_id,
            data_obj_id,
        )
        self.cur.execute(
            "INSERT INTO posts(source,metadata,data) VALUES(?,?,?)", ins_post
        )
        self.con.commit()

        post_idx = self.cur.lastrowid

        # Establish tags and links
        # Ensure the tags exist
        self.cur.executemany(
            "INSERT OR IGNORE INTO tags(tag) VALUES (?)",
            [(tag,) for tag in post.tags()],
        )
        self.con.commit()

        # Link the tags to the post
        tags = [(post_idx, tag) for tag in post.tags()]
        self.cur.executemany(
            "INSERT INTO post_tags(post, tag) VALUES(?,?)", tags)
        self.con.commit()

        # Establish any replying to

        replies = [(post_idx, replying_to)
                   for replying_to in post.replying_to()]
        self.cur.executemany(
            "INSERT INTO replies(post, replying_to) VALUES(?,?)", replies
        )
        self.con.commit()

        return None

    def new_post(self) -> core.Post:
        p = Post()
        p.bb = self
        p.set_source(self._name)
        return p

    async def len(self) -> int | Exception:
        res = self.cur.execute("SELECT COUNT(*) FROM posts")
        len_res = res.fetchone()
        return len_res[0]



import os
import sys
from vyper import v

from pathlib import Path

# I'm in a hurry and I need to match behavior between the Go API
# and the python API. I apologize to all python afficianados
# for doing go style errors.


class Error(str):
    def Error(self) -> str:
        return self


def get_project_dir() -> tuple[str, Error]:
    working_dir = os.getcwd()

    test_path = Path(working_dir)
    while test_path != Path("/"):
        if is_core_project(test_path):
            return (str(test_path), None)

        test_path = test_path.parent

    return "", Error("not a core project (or any parent directories) " + working_dir)


def is_core_project(dirpath: str) -> bool:
    # TODO: clean up this bad mix of os.path.join and Pathlib
    test_path = os.path.join(dirpath, get_core_path())
    p = Path(test_path)
    return p.is_dir()


def get_core_path() -> str:
    return "./.core"


def get_config_filename() -> str:
    return "config.yaml"


def get_config_path() -> tuple[str, Error]:
    (proj_dir, err) = get_project_dir()
    if err is not None:
        return ("", err)

    final_path = os.path.join(proj_dir, get_core_path(), get_config_filename())

    return (final_path, None)


# TODO: exception handling
def read_config(path: str):
    v.set_config_file(path)
    v.automatic_env()
    v.read_in_config()
    return v


def load_config():
    config_path, err = get_config_path()
    if err is not None:
        sys.exit(err)

    project_config = read_config(config_path)
    print("project config loaded from", config_path)
    return project_config



# TagSet implements our matching logic for message tags


from typing import Iterable


class TagSet:
    def __init__(self, tags: Iterable[str]):
        final_tags = set()
        for tag in tags:
            final_tags.add(tag)

        self.tags = final_tags

    def matches(self, incoming) -> bool:
        if len(self.tags) > len(incoming.tags):
            return False

        if len(self.tags) == 0:
            return True

        for tag in self.tags:
            if tag not in incoming.tags:
                return False
        return True

    def __str__(self):
        tag_str = ", ".join(self.tags)

        return "{" + tag_str + "}"



from typing import Iterable
import hashlib
import binascii
import os


def quick_hash(data: bytes) -> str:
    hash = hashlib.sha256(data)
    return binascii.hexlify(hash.digest()).decode()


def random_id() -> str:
    uid = os.urandom(4)
    return quick_hash(uid)


# These are interface classes and should provide no implementation


class Post:
    # Access only (set by the server)
    def index(self) -> int:
        raise NotImplementedError

    def set_index(self, idx: int):
        raise NotImplementedError

    def timestamp(self) -> float:
        raise NotImplementedError

    # Read/Write
    def source(self) -> str:
        raise NotImplementedError

    def set_source(self, src: str):
        raise NotImplementedError

    def tags(self) -> Iterable[str]:
        raise NotImplementedError

    def set_tags(self, tags: Iterable[str]):
        raise NotImplementedError

    def replying_to(self) -> Iterable[int]:
        raise NotImplementedError

    def set_replying_to(self, replying_to: Iterable[int]):
        raise NotImplementedError

    async def metadata(self) -> bytes | Exception:
        raise NotImplementedError

    def set_metadata(self, meta: bytes):
        raise NotImplementedError

    async def data(self) -> bytes | Exception:
        raise NotImplementedError

    def set_data(self, data: bytes):
        raise NotImplementedError

    def encryption(self) -> str:
        raise NotImplementedError

    def set_encryption(self, enc: str):
        raise NotImplementedError


# "Abstract" classes that define our required
# behavior for a core python transit layer
#
# These are not true python abstract base classes, but they
# will let a user know that they haven't implemented something
# that was required for what they were trying to do.
#
# Core transit layers facilitate communication between the agent and the
# blackboard.  This includes providing some basic message types that
# are tagged appropriately to represent the basics of our core multi-agent
# communication.  These are things like a "hello" "goodbye" "post", etc.
#
# NOTE: we don't bother with the interface granularity used in
# the Go implementation. this api is to solely enable *core* agents
# If a looser base class is desired, we'll need to open a issue.


class Blackboard:
    # read, write, len are the backend blackboard interface
    # these methods carry Posts to the remotely connected
    # blackboard.
    async def read(self, idx: int) -> Post | Exception:
        raise NotImplementedError

    async def write(self, post: Post) -> None | Exception:
        raise NotImplementedError

    async def len(self) -> int | Exception:
        raise NotImplementedError

    # objects are not reeeally part of the required API, but they're
    # actually probably pretty critical to effective agent environments.
    async def put_obj(self, data: bytes) -> str | Exception:
        raise NotImplementedError

    async def get_obj(self, id: str) -> bytes | Exception:
        raise NotImplementedError

    # name and set name are "required" for a CoreBlackboard
    # this is more behavioral than anything; we want connected
    # agents to self-identify as *something*
    def name(self) -> str:
        raise NotImplementedError

    def set_name(self, name: str):
        raise NotImplementedError

    # CoreBlackboards facilitate core-style multi-agent communication
    # these methods provide the basic core message types
    def message_hello(self) -> Post:
        raise NotImplementedError

    def message_goodbye(self) -> Post:
        raise NotImplementedError

    def new_post(self) -> Post:
        raise NotImplementedError



# setup.py
from setuptools import setup

with open("VERSION", "r") as f:
    version = f.read()

version = version.replace("v", "")
res = version.split("-")

# Probably a little brittle, but working for now
# Update: it was brittle
# if len(res) > 1:
#     version = res[0] + ".dev" + res[1] + "+" + res[2]
# else:
#     version = res[0].strip()

setup(version=f"{version}", test_suite="tests")



import os
import psycopg2
from . import core
from .hardcore_post import Post


class PostgreSQLTransit(core.Blackboard):
    """
    PostgreSQLTransit is a proof of concept.  It can/will be matured
    if needed.  For now, it's only made available to illustrate
    the possible forward direction.
    """

    def __init__(self, path: str):
        self._name = "pg-tester"
        # self.con = psycopg2.connect("dbname=bb user=postgres")
        self.con = psycopg2.connect(path)

        self.cur = self.con.cursor()

        print(
            "WARN: PostgreSQLTransit is not a complete implementation and will only work in narrow use cases"
        )

        # provision the db if needed
        # "index" is "id" since index is a keyword
        self.cur.execute("""
            CREATE TABLE IF NOT EXISTS posts(
                idx BIGSERIAL PRIMARY KEY,
                source TEXT,
                timestamp TIMESTAMP WITH TIME ZONE DEFAULT CURRENT_TIMESTAMP,
                metadata TEXT,
                data TEXT,
                encryption TEXT DEFAULT 'aes-gcm'
            );
            """)
        self.con.commit()

        self.cur.execute("CREATE TABLE IF NOT EXISTS tags(tag TEXT UNIQUE);")
        self.con.commit()

        self.cur.execute("""
            CREATE TABLE IF NOT EXISTS post_tags(
                post BIGINT REFERENCES posts(idx),
                tag TEXT REFERENCES tags(tag)
            );
            """)
        self.con.commit()

        self.cur.execute("""
            CREATE TABLE IF NOT EXISTS replies(
                post BIGINT REFERENCES posts(idx),
                replying_to BIGINT REFERENCES posts(idx)
            );
            """)
        self.con.commit()

        # Ensure we can use a directory next to the DB for an object store
        db_dir = os.path.dirname(path)
        self.obj_dir = os.path.join(db_dir, "obj")
        try:
            os.makedirs(self.obj_dir)
        except:
            pass

    async def put_obj(self, data: bytes) -> str | Exception:
        obj_id = core.random_id()
        obj_path = os.path.join(self.obj_dir, obj_id)
        with open(obj_path, "wb") as f:
            f.write(data)
        return obj_id

    async def get_obj(self, id: str) -> bytes | Exception:
        obj_path = os.path.join(self.obj_dir, id)
        with open(obj_path, "rb") as f:
            return f.read()

    # NOTE: Agents are expecting to index from zero, but PostgreSQLite starts with row idx 1
    async def read(self, idx: int) -> core.Post | Exception:
        adjusted_idx = idx + 1
        self.cur.execute(
            "SELECT idx,source,metadata,data FROM posts WHERE idx = %s;",
            (adjusted_idx,),
        )
        self.con.commit()

        post_data = self.cur.fetchone()  # This should be unique
        if post_data is None:
            return RuntimeError(f"idx {adjusted_idx} not found")

        # Query the tags associated with the post
        self.cur.execute("SELECT tag FROM post_tags WHERE post = %s;", (adjusted_idx,))
        self.con.commit()

        tags = self.cur.fetchall()  # This should be unique
        tag_list = [x[0] for x in tags]

        # argument for: agents should never really be dealing with those indices directly
        # against: we expose the index to the user via the post.index() method, if this
        # is not internally consistent, it will make for bad behavior.
        post = self.new_post()
        # this will return an index to the agent that they could use to 'read(idx)' to get the message back.
        post.set_index(post_data[0] - 1)
        post.set_source(post_data[1])
        post.set_tags(tag_list)
        post.set_metadata(post_data[2].encode())
        post.set_data(post_data[3].encode())

        # temporary kluge
        post._m_retrieved = False
        post._d_retrieved = False

        return post

    async def write(self, post: core.Post) -> None | Exception:
        # Indirect our data onto disk

        m_data = await post.metadata()
        if isinstance(m_data, Exception):
            return m_data

        metadata_obj_id = ""
        if len(m_data) > 0:
            metadata_obj_id = await self.put_obj(m_data)
            if isinstance(metadata_obj_id, Exception):
                return metadata_obj_id

        data = await post.data()
        if isinstance(data, Exception):
            return data

        data_obj_id = ""
        if len(data) > 0:
            data_obj_id = await self.put_obj(data)
            if isinstance(data_obj_id, Exception):
                return data_obj_id
            post.set_data(data_obj_id.encode())

        ins_post = (
            post.source(),
            metadata_obj_id,
            data_obj_id,
        )
        self.cur.execute(
            "INSERT INTO posts(source,metadata,data) VALUES(%s,%s,%s) RETURNING idx;",
            ins_post,
        )
        post_idx = self.cur.fetchone()
        self.con.commit()

        # Establish tags and links
        # Ensure the tags exist
        self.cur.executemany(
            "INSERT INTO tags(tag) VALUES (%s) ON CONFLICT DO NOTHING;",
            [(tag,) for tag in post.tags()],
        )
        self.con.commit()

        # Link the tags to the post
        tags = [(post_idx, tag) for tag in post.tags()]
        self.cur.executemany("INSERT INTO post_tags(post, tag) VALUES(%s,%s);", tags)
        self.con.commit()

        # Establish any replying to

        replies = [(post_idx, replying_to) for replying_to in post.replying_to()]
        self.cur.executemany(
            "INSERT INTO replies(post, replying_to) VALUES(%s,%s);", replies
        )
        self.con.commit()

        return None

    def new_post(self) -> core.Post:
        p = Post()
        p.bb = self
        p.set_source(self._name)
        return p

    async def len(self) -> int | Exception:
        self.cur.execute("SELECT COUNT(*) FROM posts;")
        len_res = self.cur.fetchone()
        return len_res[0]
